# Copyright 2020 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for
# Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import cdflib
import logging
import re
import numpy as np
import xarray as xr
import datetime
from datetime import timedelta
from pyspedas.tplot_tools import store_data
from pyspedas.tplot_tools import tplot
from pyspedas.tplot_tools import options
import pyspedas
import copy
from collections.abc import Iterable


def cdf_to_tplot(filenames, mastercdf=None, varformat=None, exclude_format=None, get_support_data=False, get_metadata=False,
                 get_ignore_data=False, string_encoding='ascii',
                 prefix='', suffix='', plot=False, merge=False,
                 center_measurement=False, notplot=False, varnames=None):
    """
    This function will automatically create tplot variables from CDF files.  In general, the files should be
    ISTP compliant for this importer to work.  Each variable is read into a new tplot variable (a.k.a an xarray DataArray),
    and all associated file/variable metadata is read into the attrs dictionary.

    .. note::
        Variables must have an attribute named "VAR_TYPE". If the attribute entry
        is "data" (or "support_data"), then they will be added as tplot variables.
        Additionally, data variables should have attributes named "DEPEND_TIME" or
        "DEPEND_0" that describes which variable is x axis.  If the data is 2D,
        then an attribute "DEPEND_1" must describe which variable contains the
        secondary axis.

    Parameters:
        filenames : str/list of str
            The file names and full paths of CDF files.
        mastercdf : str
            The file name of a master CDF to be used, if any
        varformat : str or list[str]
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.
        exclude_format : str or list[str]
            The file variable formats to exclude from loading into tplot.  Wildcard character
            "*" is accepted. By default, no variables are excluded.
        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".
        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
        suffix: str
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.
        plot: bool
            The data is plotted immediately after being generated.  All tplot
            variables generated from this function will be on the same plot.
        merge: bool
            If True, then data from different cdf files will be merged into
            a single tplot variable.
        get_ignore_data: bool
            Data with an attribute "VAR_TYPE" with a value of "ignore_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".
        center_measurement: bool
            If True, the CDF epoch variables are time-shifted to the middle
            of the accumulation interval by their DELTA_PLUS_VAR and
            DELTA_MINUS_VAR variable attributes
        notplot: bool
            If True, then data are returned in a hash table instead of
            being stored in tplot variables (useful for debugging, and
            access to multi-dimensional data products)
        varnames: str or list of str
            Load these variables only. If None or [] or ['*'], then load everything.

    Returns:
        List of tplot variables created (unless notplot keyword is used).
    """

    stored_variables = []
    epoch_cache = {}
    output_table = {}
    metadata = {}

    new_cdflib = False
    if cdflib.__version__ > "0.4.9":
        new_cdflib = True
        logging.debug("Using new version of cdflib (%s)", cdflib.__version__)
    else:
        new_cdflib = False
        logging.debug("Using old version of cdflib (%s)", cdflib.__version__)

    if prefix is None:
        prefix = ''

    if suffix is None:
        suffix = ''

    # When doing varname, varformat, or exclude_format checks, do we need to take prefixes or suffixes into account?
    check_pre_suff = False
    if prefix != '' or suffix != '':
        check_pre_suff = True

    # If nontrivial varnames or varformat are explicitly supplied, load the requested data whether or not
    # it's marked as support_data in the CDF.

    if varformat is not None:
        get_support_data = True

    if varnames is None:
        varnames = []

    if not isinstance(varnames, list):
        varnames = [varnames]

    if len(varnames) > 0:
        get_support_data = True
        if '*' in varnames:
            varnames = []

    # pyspedas.tplot_tools.data_quants = {}
    if isinstance(filenames, str):
        filenames = [filenames]
    elif isinstance(filenames, list):
        pass
    else:
        logging.warning("Invalid filenames input. Must be string or list of strings.")
        return stored_variables

    var_type = ['data']
    if varformat is None:
        varformat = ".*"
    if get_support_data:
        var_type.append('support_data')
    if get_metadata:
        var_type.append('metadata')
    if get_ignore_data:
        var_type.append('ignore_data')

    # Replace lists with regex alternation
    if isinstance(varformat, list):
        varformat = '|'.join(varformat)
    # Replace spaces with regex alternation
    if ' ' in varformat:
        explode=varformat.split()
        varformat='|'.join(explode)

    varformat = varformat.replace("*", ".*")
    var_regex = re.compile(varformat)

    if exclude_format is not None:
        # Replace spaces and lists with regex alternation as we did for varformat
        if isinstance(exclude_format, list):
            exclude_format = '|'.join(exclude_format)
        # Replace spaces with regex alternation
        if ' ' in exclude_format:
            explode=exclude_format.split()
            exclude_format='|'.join(explode)

        exclude_format = exclude_format.replace("*",".*")
        exclude_regex = re.compile(exclude_format)
    else:
        exclude_regex = None


    # This step may not be appropriate if the lexicographic sort does not correspond to a time sort. (For example,
    # if filenames contain orbit numbers rather than dates, and no leading zeroes are used.)  JWL 2023-03-17

    filenames.sort()

    # Get metadata from master CDF, if provided
    # In IDL, cdf2tplot uses the first file provided as a de-facto master CDF.
    # in pyspedas, cdf_to_tplot can do things like loading data from all 4 MMS probes in a single call.
    # So, we can't always use the first CDF in the list, because it may not apply to other files in the list.
    # Therefore, we supply a master CDF, if needed, in a separate argument. JWL 2023-03-17

    if not mastercdf is None:
        mastercdf_flag = True
        logging.debug('Processing master CDF %s', mastercdf)
        master_cdf_file = cdflib.CDF(mastercdf)
        master_cdf_file.string_encoding = string_encoding
        master_cdf_info = master_cdf_file.cdf_info()
        if new_cdflib:
            master_cdf_variables = master_cdf_info.rVariables + master_cdf_info.zVariables
        else:
            master_cdf_variables = master_cdf_info['rVariables'] + master_cdf_info['zVariables']

        logging.debug("master_cdf_variables: " + str(master_cdf_variables))
    else:
        mastercdf_flag = False

    logging.debug("Input filenames: " + str(filenames))
    for filename in filenames:
        logging.debug('Processing filename %s', filename)
        cdf_file = cdflib.CDF(filename)
        cdf_file.string_encoding = string_encoding
        cdf_info = cdf_file.cdf_info()
        if new_cdflib:
            all_cdf_variables = cdf_info.rVariables + cdf_info.zVariables
        else:
            all_cdf_variables = cdf_info['rVariables'] + cdf_info['zVariables']

        logging.debug("all_cdf_variables: " + str(all_cdf_variables))
        if not mastercdf_flag:
            # If not using a master CDF, each CDF is its own master
            master_cdf_file = cdf_file
            mastercdf = filename
            master_cdf_variables = all_cdf_variables
        # User defined variables.
        if len(varnames) > 0:
            load_cdf_variables = [value for value in varnames if value in all_cdf_variables]
            if check_pre_suff:
                pre_suff = [value for value in all_cdf_variables if prefix+value+suffix in varnames]
                load_cdf_variables.extend(pre_suff)
        else:
            load_cdf_variables = all_cdf_variables

        try:
            gatt = master_cdf_file.globalattsget()
        except:
            logging.warning('Unable to get global attributes for filename %s', mastercdf)
            gatt = {}

        for var in load_cdf_variables:
            if not re.match(var_regex, var) and (not check_pre_suff or not re.match(var_regex, prefix+var+suffix)):
                logging.debug("Variable %s does not match varformat, skipping", var)
                continue
            elif exclude_regex is not None and (re.match(exclude_regex, var) or (check_pre_suff and re.match(exclude_regex, prefix+var+suffix))):
                logging.debug("Variable %s matches exclude_format, skipping", var)
                continue
            logging.debug('Processing variable attributes for %s', var)
            try:
                var_atts = master_cdf_file.varattsget(var)
            except ValueError:
                logging.warning("Unable to get variable attributes for %s in file %s, skipping", var, mastercdf)
                continue

            if 'VIRTUAL' in var_atts:
                this_virtual = var_atts['VIRTUAL'].lower()
                if this_virtual=="true":
                    logging.debug("Skipping virtual variable %s",var)
                    continue
            elif 'FUNCT' in var_atts and 'COMPONENT_0' in var_atts:
                logging.info("Variable %s not marked as VIRTUAL, but has FUNCT and COMPONENT_0 attributes; skipping", var)
                continue

            if 'VAR_TYPE' in var_atts:
                this_var_type = var_atts['VAR_TYPE'].lower()
            elif 'PARAMETER_TYPE' in var_atts:
                this_var_type = var_atts['PARAMETER_TYPE'].lower()
            else:
                # 'VAR_TYPE' and 'PARAMETER_TYPE' not found in the variable attributes
                logging.info('No VAR_TYPE or PARAMETER_TYPE attributes defined for variable %s, skipping', var)
                continue

            if this_var_type in var_type:
                var_properties = master_cdf_file.varinq(var)
                var_properties_data_cdf = cdf_file.varinq(var)

                # Find data name and if it is already in stored variables
                if 'TPLOT_NAME' in var_atts:
                    var_name = prefix + var_atts['TPLOT_NAME'] + suffix
                else:
                    var_name = prefix + var + suffix

                # Is this variable marked as non-record-varying?  This may
                # differ between the data and master CDFs.

                if new_cdflib:
                    rec_vary_data \
                        = var_properties_data_cdf.Rec_Vary
                    rec_vary_master \
                        = var_properties.Rec_Vary
                else:
                    rec_vary_data \
                        = var_properties_data_cdf["Rec_Vary"]
                    rec_vary_master \
                        = var_properties["Rec_Vary"]
                if (rec_vary_master != rec_vary_data):
                    logging.warning("Master and data CDFs have different values for Rec_Vary property on variable %s, using %s from master CDF.", var, rec_vary_master)
                rec_vary = rec_vary_master

                nrv_has_times = False
                if "DEPEND_TIME" in var_atts:
                    x_axis_var = var_atts["DEPEND_TIME"]
                    if not rec_vary:
                        logging.warning("Variable %s is marked non-record-varying, but has DEPEND_TIME attribute",var)
                        nrv_has_times = True
                elif "DEPEND_0" in var_atts:
                    x_axis_var = var_atts["DEPEND_0"]
                    if not rec_vary:
                        logging.warning("Variable %s is marked non-record-varying, but has DEPEND_0 attribute",var)
                        nrv_has_times = True

                else:
                    # non-record varying variables (NRVs)
                    # added by egrimes, 13Jan2021
                    # here we assume if there isn't a DEPEND_TIME or DEPEND_0, there are no other depends
                    logging.debug(
                        'No DEPEND_TIME or DEPEND_0 attributes found for variable %s, filename %s . Treating as non-record-varying.',
                        var, filename)
                    if rec_vary and ('epoch' not in var.lower()):
                        logging.warning("Variable %s is marked as record-varying, but no DEPEND_TIME or DEPEND_0 attributes found. Treating as non-record-varying.",var)
                    try:
                        ydata = cdf_file.varget(var)
                    except:
                        logging.debug('Unable to get ydata for non-record-varying variable %s, filename %s', var, filename)
                        continue

                    if ydata is None:
                        continue

                    # since NRVs don't vary with time, they shouldn't vary across files
                    output_table[var_name] = {'y': ydata}

                    continue

                if x_axis_var not in master_cdf_variables:
                    logging.warning('Variable %s timestamp variable %s not found, skipping', var, x_axis_var)
                    continue

                if new_cdflib:
                    data_type_description \
                        = cdf_file.varinq(x_axis_var).Data_Type_Description
                else:
                    data_type_description \
                        = cdf_file.varinq(x_axis_var)["Data_Type_Description"]


                if epoch_cache.get(filename + x_axis_var) is None:
                    delta_plus_var = 0.0
                    delta_minus_var = 0.0
                    delta_time = 0.0

                    # Skip variables with ValueErrors.
                    try:
                        xdata = cdf_file.varget(x_axis_var)
                        epoch_var_atts = cdf_file.varattsget(x_axis_var)
                    except ValueError:
                        logging.debug('Problem getting data for variable %s, filename %s', var, filename)
                        continue

                    # check for DELTA_PLUS_VAR/DELTA_MINUS_VAR attributes
                    if center_measurement:
                        if 'DELTA_PLUS_VAR' in epoch_var_atts:
                            delta_plus_var = cdf_file.varget(epoch_var_atts['DELTA_PLUS_VAR'])
                            delta_plus_var_att = cdf_file.varattsget(epoch_var_atts['DELTA_PLUS_VAR'])

                            # check if a conversion to seconds is required
                            if 'SI_CONVERSION' in delta_plus_var_att:
                                si_conv = delta_plus_var_att['SI_CONVERSION']
                                delta_plus_var = delta_plus_var.astype(float) * np.float64(si_conv.split('>')[0])
                            elif 'SI_CONV' in delta_plus_var_att:
                                si_conv = delta_plus_var_att['SI_CONV']
                                delta_plus_var = delta_plus_var.astype(float) * np.float64(si_conv.split('>')[0])

                        if 'DELTA_MINUS_VAR' in epoch_var_atts:
                            delta_minus_var = cdf_file.varget(epoch_var_atts['DELTA_MINUS_VAR'])
                            delta_minus_var_att = cdf_file.varattsget(epoch_var_atts['DELTA_MINUS_VAR'])

                            # check if a conversion to seconds is required
                            if 'SI_CONVERSION' in delta_minus_var_att:
                                si_conv = delta_minus_var_att['SI_CONVERSION']
                                delta_minus_var = delta_minus_var.astype(float) * np.float64(si_conv.split('>')[0])
                            elif 'SI_CONV' in delta_minus_var_att:
                                si_conv = delta_minus_var_att['SI_CONV']
                                delta_minus_var = delta_minus_var.astype(float) * np.float64(si_conv.split('>')[0])

                        # sometimes these are specified as arrays
                        if isinstance(delta_plus_var, np.ndarray) and isinstance(delta_minus_var, np.ndarray):
                            delta_time = (delta_plus_var - delta_minus_var) / 2.0
                        else:  # and sometimes constants
                            if delta_plus_var != 0.0 or delta_minus_var != 0.0:
                                delta_time = (delta_plus_var - delta_minus_var) / 2.0

                if epoch_cache.get(filename + x_axis_var) is None:
                    if ('CDF_TIME' in data_type_description) or \
                            ('CDF_EPOCH' in data_type_description):
                        # the old way:
                        # store the times as unix times, and cache them
                        # xdata = cdfepoch.unixtime(xdata)
                        # epoch_cache[filename+x_axis_var] = np.array(xdata)+delta_time
                        # the new way:
                        # store and cache the datetime objects directly
                        # and delay conversion to unix times until get_data is called

                        # Cluster uses multidimensional DEPEND_0 values on some "caveat" and "dsettings" variables. This will cause the
                        # xdata[0] < 0.0 test to crash
                        if len(xdata.shape) > 1:
                            logging.warning("CDF DEPEND_0 attribute %s for variable %s is multidimensional with shape %s, skipping", var_atts['DEPEND_0'], var, str(xdata.shape),)
                            continue
                        # Cluster apparently uses (-1.0e-31) as time tag fill values??  Better check...
                        # if xdata[0] < 0.0:
                        #    logging.warning("CDF time tag %e for variable %s cannot be converted to datetime, skipping",xdata[0],var)
                        #    continue

                        # Check for all-fill Cluster times
                        # NOTE: At least for Cluster onboard moments downloaded from CSA,
                        # there is a FILLVAL specified on the time_tags variable, but
                        # it doesn't seem to match the actual fill value we're seeing.
                        # The time_tags FILLVAL attribute has -1e+31 (very large negative value) while
                        # the ones we're actually seeing in the data are -1e-31 (very small negative value)
                        is_cluster_fill = xdata == -1.0e-31
                        if is_cluster_fill.all():
                            logging.warning("Time variable %s for data variable %s has values all equal to -1.0e-31", x_axis_var, var)
                        # Check if time variable has a FILLVAL attribute defined.
                        if 'FILLVAL' in epoch_var_atts:
                            fillval = epoch_var_atts['FILLVAL']
                            is_fillval = xdata == fillval
                            if is_fillval.any():
                                logging.warning("Time variable %s for data variable %s contain at least one time equal to FILLVAL (%e)", x_axis_var, var, fillval)
                        xdata = np.array(cdflib.cdfepoch.to_datetime(xdata))
                        if isinstance(xdata[0],datetime.datetime):
                            # old cdflib < 1.0.0 returns datetime.datetime objects
                            if isinstance(delta_time, np.ndarray) or isinstance(delta_time, list):
                                delta_t = np.array([timedelta(seconds=dtime) for dtime in delta_time])
                            else:
                                delta_t = timedelta(seconds=delta_time)
                        else:
                            # new cdflib >= 1.0.0 returns np.datetime64 objects
                            if isinstance(delta_time, np.ndarray) or isinstance(delta_time, list):
                                delta_t = np.array([np.timedelta64(int(dtime*1e9),'ns') for dtime in delta_time])
                            else:
                                delta_t = np.timedelta64(int(delta_time*1e9),'ns')

                        epoch_cache[filename + x_axis_var] = xdata + delta_t
                else:
                    xdata = epoch_cache[filename + x_axis_var]

                try:
                    ydata = cdf_file.varget(var)
                except:
                    logging.warning('Unable to get ydata for variable %s', var)
                    continue

                if ydata is None:
                    logging.info('No ydata for variable %s', var)
                    continue
                elif np.isscalar(ydata):
                    # Cluster sets FILLVAL attributes on scalar quantities (!) so we need to check...
                    # This can happen for density variables in the Cluster onboard moments loaded from CSA.
                    # It may be due to no valid data being available, but CSA makes a CDF with a single
                    # time value and data point that are both fillvals.
                    logging.info('ydata for variable %s is a scalar, converting to numpy array',var)
                    # We won't worry here about how many dimensions it's supposed to have.  We'll fix that below if needed.  For now, we just want to be
                    # sure it's not a scalar.
                    ydata = np.array(ydata)
                if "FILLVAL" in var_atts:
                    if new_cdflib:
                        thisvar_dtd = var_properties.Data_Type_Description
                    else:
                        thisvar_dtd = var_properties["Data_Type_Description"]

                    fillval = var_atts['FILLVAL']

                    if (thisvar_dtd == 'CDF_FLOAT' or
                            thisvar_dtd == 'CDF_REAL4' or
                            thisvar_dtd == 'CDF_DOUBLE' or
                            thisvar_dtd == 'CDF_REAL8'):

                        is_fill_cond = ydata == fillval
                        if is_fill_cond.all():
                            logging.warning("Floating point data values for variable %s are all fillval (%e)",var, fillval)
                            ydata[is_fill_cond] = np.nan
                        elif is_fill_cond.any():
                            ydata[is_fill_cond] = np.nan
                        else:
                            # No fillvals, nothing to do
                            pass
                    elif thisvar_dtd[:7] == 'CDF_INT':
                        # NaN is only valid for floating point data
                        # but we still need to handle FILLVAL's for
                        # integer data, so we'll just set those to 0
                        is_fill_cond = ydata == fillval
                        if is_fill_cond.all():
                            logging.warning("Integer data values for variable %s are all fillval (%d).",var, fillval)
                            ydata[is_fill_cond] = 0
                        elif is_fill_cond.any():
                            ydata[is_fill_cond] = 0
                        else:
                            # No fillvals, nothing to do
                            pass

                # Check dimensions of ydata to see if a leading time dimension has been lost
                # This seems to happen with some Cluster CDFs, at least the ones served by CSA,
                # if a variable only has a single timestamp. For example, the CP_CIS-HIA_ONBOARD_MOMENTS datatype, as seen
                # in test_load_csa_mom_data.

                num_times = len(xdata)
                ydims = ydata.shape
                y_ndims = len(ydata.shape)
                if num_times == 1:
                    if y_ndims == 0:
                        logging.warning("Restoring missing time dimension for scalar-valued variable %s", var)
                        ydata = ydata.reshape(1)
                        ydims = ydata.shape
                        y_ndims = len(ydata.shape)
                    elif ydims[0] != 1:
                        logging.warning("Restoring missing time dimension for array-valued variable %s", var)
                        ydata = ydata.reshape(1,*ydims)
                        ydims = ydata.shape
                        y_ndims = len(ydata.shape)
                elif nrv_has_times and (num_times > 2) and (ydims[0] != num_times):
                    # This case is primarily to catch some MMS FEEPS support variables
                    # that's marked NRV, but has a DEPEND_0.  Here, we ignore the times,
                    # make the tplot variable from just the Y data, and skip the rest of
                    # the metadata processing for this variable.
                    logging.warning("Ignoring times for probably non-record-varying variable %s", var_name)
                    output_table[var_name] = {'y': ydata}
                    continue

                tplot_data = {'x': xdata, 'y': ydata}

                # We want to know if this is a spectrogram or not.  If not, don't make
                # "v", "v1", "v2" entries. Technically, a vector quantity should have a DEPEND_1
                # with numeric values.  Most of the time, they are provided as strings, or simply
                # omitted.  But if we made a "v" variable for it, it would break a lot of code
                # that only expects to get (times, data) back from a get_data call on a non-spectral variable.

                is_spectrogram = False
                if 'DISPLAY_TYPE' in var_atts:
                    disp_type = var_atts['DISPLAY_TYPE'].lower()
                    if "spect" in disp_type:
                        is_spectrogram = True

                # Data may depend on other data in the CDF.
                depend_1 = None
                depend_2 = None
                depend_3 = None
                if "DEPEND_1" in var_atts:
                    if y_ndims < 2:
                        logging.warning("Variable %s has only %d dimension (including time), but has a DEPEND_1 attribute. Removing attribute.", var, y_ndims)
                        depend_1 = None
                    elif var_atts["DEPEND_1"] in master_cdf_variables:
                        try:
                            # Check for correct shape, matching time and data dimensions
                            dep_name = var_atts["DEPEND_1"]
                            depend_1 = np.array(master_cdf_file.varget(dep_name))
                            # String-valued DEPEND_1 handling
                            # This is not strictly ISTP compliant, but it's extremeley common. For
                            # example, vector-valued data will often have DEPEND_1 values that are more like
                            # labels, i.e. ['x', 'y', 'z']
                            # Previous versions of cdf_to_tplot simply ignored string-valued DEPEND_N.
                            # But some missions (e.g. ERG) really do want string values as 'v1' or 'v2' tags.
                            #
                            # The situation is further complicated by the fact that cdflib seems to introduce
                            # extra leading or trailing dimensions on string-valued variables.
                            #
                            # As of cdflib version 1.3.3, it seems the issue has been corrected.  I will leave
                            # the test and warning enabled, in case of a regression later.  JWL 2025-04-04
                            #
                            if depend_1 is not None and depend_1.dtype.type is np.str_:
                                # Get the original array dimensions from the variable properties
                                dp_props = master_cdf_file.varinq(dep_name)
                                if new_cdflib:
                                    orig_dimensions = dp_props.Dim_Sizes
                                else:
                                    orig_dimensions = dp_props['Dim_Sizes']
                                reshape_dim = tuple(orig_dimensions)
                                if depend_1.shape != reshape_dim:
                                    logging.warning('Variable %s has shape %s. Its DEPEND_1 attribute %s is string-valued, and cdflib returned dimensions %s which do not match original dimensions %s.',var, ydata.shape, dep_name, depend_1.shape, reshape_dim)
                                    if len(depend_1.shape) == 1 and depend_1.shape[0] == ydata.shape[1]:
                                        logging.warning('Returned dimensions are a better match than original, no reshaping required')
                                    elif len(reshape_dim) == 1 and reshape_dim[0] == ydata.shape[1]:
                                        logging.warning('Original dimensions are a better match to data dimensions, reshaping.  Updating to a more recent version of cdflib may get rid of this warning.')
                                        depend_1 = np.reshape(depend_1, reshape_dim)
                                    else:
                                        logging.warning('Neither original nor returned dimensions match data shape, ignoring this DEPEND_1.')
                                        depend_1 = None
                                pass
                            dep_dims = depend_1.shape
                            dep_ndims = len(dep_dims)
                            if dep_ndims == 0:
                                logging.warning("Variable %s DEPEND_1 attribute %s is zero-dimensional, Removing attribute.", var, dep_name)
                                depend_1 = None
                            elif dep_ndims == 1:
                                # Not time varying
                                if dep_dims[0] != ydims[1]:
                                    logging.warning("Variable %s DEPEND_1 attribute %s has length %d, but corresponding data dimension has length %d. Removing attribute.",var,dep_name,dep_dims[0],ydims[1])
                                    depend_1 = None
                            elif dep_ndims == 2:
                                # time-varying or otherwise multidimensional
                                if dep_dims[0] != num_times:
                                    logging.warning("Variable %s is 2-dimensional, but first dimension of DEPEND_1 attribute %s has size %d versus num_times %d. Attribute will be kept (for now).",var,dep_name,dep_dims[0], num_times)
                                    if dep_dims[0] == 1 and dep_dims[1] == ydims[1]:
                                        # RBSP EMPHISIS HFR_Spectra has this
                                        logging.warning("Variable %s DEPEND_1 attribute %s has dimensions 1 x y_dims[1]; reshaping to 1-D array.", var, dep_name)
                                        depend_1 = np.reshape(depend_1, (ydims[1],))
                                    # Or, it could be ERG HEP omniflux data with an extra (length-2) dimension as upper/lower bounds.
                                    # So for now, we'll allow it.
                                    pass
                                if dep_dims[1] != ydims[1]:
                                    # ERG XEP seems to make a 9x2 rather than a 2x9 array
                                    logging.warning("Variable %s is 2-dimensional, but second dimension of DEPEND_1 attribute %s has data length %d, but corresponding data dimension has length %d. Attribute will be kept (for now).",var,dep_name,dep_dims[1],ydims[1])
                                    #depend_1 = None
                                    pass
                            else:
                                # Too many dimensions
                                # ERG LEPE has time dependent DEPEND_1 with an extra dimension for upper/lower limits, so
                                # we need to allow this for now, or at least add a flag to skip this check.
                                logging.warning("Variable %s DEPEND_1 attribute %s has too many dimensions (%d). Keeping extra dimensions (for now).",var,dep_name,dep_ndims)
                                #depend_1 = None
                                pass

                        except ValueError:
                            logging.warning('Unable to get DEPEND_1 variable %s while processing %s',
                                            var_atts["DEPEND_1"], var)
                            pass
                if "DEPEND_2" in var_atts:
                    if y_ndims < 3:
                        logging.warning("Variable %s has only %d dimensions (including time), but has a DEPEND_2 attribute. Removing attribute.", var, y_ndims)
                        depend_2 = None
                    elif var_atts["DEPEND_2"] in master_cdf_variables:

                        try:
                            # Check for correct shape, matching time and data dimensions
                            dep_name = var_atts["DEPEND_2"]
                            depend_2 = np.array(master_cdf_file.varget(dep_name))
                            # String-valued DEPEND_N handling
                            # This is not strictly ISTP compliant, but it's extremeley common. For
                            # example, vector-valued data will often have DEPEND_1 values that are more like
                            # labels, i.e. ['x', 'y', 'z']
                            # Previous versions of cdf_to_tplot simply ignored string-valued DEPEND_N.
                            # But some missions (e.g. ERG) really do want string values as 'v1' or 'v2' tags.
                            #
                            # The situation is further complicated by the fact that cdflib seems to introduce
                            # extra leading or trailing dimensions on string-valued variables.
                            if depend_2 is not None and depend_2.dtype.type is np.str_:
                                # Get the original array dimensions from the variable properties
                                dp_props = master_cdf_file.varinq(dep_name)
                                if new_cdflib:
                                    orig_dimensions = dp_props.Dim_Sizes
                                else:
                                    orig_dimensions = dp_props['Dim_Sizes']
                                reshape_dim = tuple(orig_dimensions)
                                depend_2 = np.reshape(depend_2, reshape_dim)
                                #depend_2 = None
                                pass

                            dep_dims = depend_2.shape
                            dep_ndims = len(dep_dims)
                            if dep_ndims == 0:
                                logging.warning("Variable %s DEPEND_2 attribute %s is zero-dimensional. Removing attribute.", var,
                                                dep_name)
                                depend_2 = None
                            elif dep_ndims == 1:
                                # Not time varying
                                if dep_dims[0] != ydims[2]:
                                    logging.warning(
                                        "Variable %s DEPEND_2 attribute %s has length %d, but corresponding data dimension has length %d. Removing attribute.",
                                        var, dep_name, dep_dims[0], ydims[2])
                                    depend_2 = None
                            elif dep_ndims == 2:
                                # time-varying or otherwise multidimensional
                                if dep_dims[0] != num_times:
                                    logging.warning(
                                        "Variable %s multidimensional DEPEND_2 attribute %s has %d elements in first dimension, but data has %d times. Removing attribute.",
                                        var, dep_name, dep_dims[0], num_times)
                                    depend_2 = None
                                if dep_dims[1] != ydims[2]:
                                    logging.warning(
                                        "Variable %s multidimensional DEPEND_2 attribute %s has %d elements in second dimension, but corresponding data dimension has length %d. Removing attribute.",
                                        var, dep_name, dep_dims[1], ydims[2])
                                    depend_2 = None
                            else:
                                # Too many dimensions
                                logging.warning(
                                    "Variable %s DEPEND_2 attribute %s has too many dimensions (%d). Removing attribute.",
                                    var, dep_name, dep_ndims)
                                depend_2 = None
                        except ValueError:
                            logging.warning('Unable to get DEPEND_2 variable %s while processing %s',
                                            var_atts["DEPEND_2"], var)
                if "DEPEND_3" in var_atts:
                    if y_ndims < 4:
                        # TWINS imager data has this
                        logging.warning("Variable %s has only %d dimensions (including time), but has a DEPEND_3 attribute. Removing attribute.", var, y_ndims)
                        depend_3 = None
                    elif var_atts["DEPEND_3"] in master_cdf_variables:
                        try:
                            # Check for correct shape, matching time and data dimensions
                            dep_name = var_atts["DEPEND_3"]
                            depend_3 = np.array(master_cdf_file.varget(dep_name))
                            # String-valued DEPEND_N handling
                            # This is not strictly ISTP compliant, but it's extremeley common. For
                            # example, vector-valued data will often have DEPEND_1 values that are more like
                            # labels, i.e. ['x', 'y', 'z']
                            # Previous versions of cdf_to_tplot simply ignored string-valued DEPEND_N.
                            # But some missions (e.g. ERG) really do want string values as 'v1' or 'v2' tags.
                            #
                            # The situation is further complicated by the fact that cdflib seems to introduce
                            # extra leading or trailing dimensions on string-valued variables.
                            if depend_3 is not None and depend_3.dtype.type is np.str_:
                                # Get the original array dimensions from the variable properties
                                dp_props = master_cdf_file.varinq(dep_name)
                                if new_cdflib:
                                    orig_dimensions = dp_props.Dim_Sizes
                                else:
                                    orig_dimensions = dp_props['Dim_Sizes']
                                reshape_dim = tuple(orig_dimensions)
                                depend_3 = np.reshape(depend_3, reshape_dim)
                                #depend_3 = None
                                pass

                            dep_dims = depend_3.shape
                            dep_ndims = len(dep_dims)
                            if dep_ndims == 0:
                                logging.warning("Variable %s DEPEND_3 attribute %s is zero-dimensional. Removing attribute.", var,
                                                dep_name)
                                depend_3 = None
                            elif dep_ndims == 1:
                                # Not time varying
                                if dep_dims[0] != ydims[3]:
                                    logging.warning(
                                        "Variable %s DEPEND_3 attribute %s has length %d, but corresponding data dimension has length %d. Removing attribute.",
                                        var, dep_name, dep_dims[0], ydims[3])
                                    depend_3 = None
                            elif dep_ndims == 2:
                                # time-varying, or otherwise multidimensional
                                if dep_dims[0] != num_times:
                                    logging.warning(
                                        "Variable %s multidimensional DEPEND_3 attribute %s has %d elements in first dimension, but data has %d times. Removing attribute.",
                                        var, dep_name, dep_dims[0], num_times)
                                    depend_3 = None
                                if dep_dims[1] != ydims[3]:
                                    logging.warning(
                                        "Variable %s multidimensional DEPEND_3 attribute %s has %d elements in second dimension, but corresponding data dimension has length %d. Removing attribute.",
                                        var, dep_name, dep_dims[1], ydims[3])
                                    depend_3 = None
                            else:
                                # Too many dimensions
                                logging.warning(
                                    "Variable %s DEPEND_3 attribute %s has too many dimensions (%d). Removing attribute.",
                                    var, dep_name, dep_ndims)
                                depend_3 = None
                        except ValueError:
                            logging.warning('Unable to get DEPEND_3 variable %s while processing %s',
                                            var_atts["DEPEND_3"], var)

                nontime_varying_depends = []

                # Fill in any missing depend_n values (skipping this for now)
                ndims = len(ydata.shape)
                if ndims >= 2 and depend_1 is None:
                    # This is so common, we won't bother logging it
                    # depend_1 = np.arange(ydata.shape[1])
                    pass
                if ndims >= 3 and depend_2 is None:
                    #logging.warning("Variable %s has %d dimensions, but no DEPEND_2, adding index range for dimension 2", var_name, ndims )
                    #depend_2 = np.arange(ydata.shape[2])
                    pass
                if ndims >= 4 and depend_3 is None:
                    #logging.warning("Variable %s has %d dimensions, but no DEPEND_3, adding index range for dimension 3", var_name, ndims )
                    #depend_3 = np.arange(ydata.shape[3])
                    pass

                if depend_1 is not None and depend_2 is not None and depend_3 is not None:
                    tplot_data['v1'] = depend_1
                    tplot_data['v2'] = depend_2
                    tplot_data['v3'] = depend_3

                    if len(depend_1.shape) == 1:
                        nontime_varying_depends.append('v1')
                    if len(depend_2.shape) == 1:
                        nontime_varying_depends.append('v2')
                    if len(depend_3.shape) == 1:
                        nontime_varying_depends.append('v3')

                elif depend_1 is not None and depend_2 is not None:
                    tplot_data['v1'] = depend_1
                    tplot_data['v2'] = depend_2
                    if len(depend_1.shape) == 1:
                        nontime_varying_depends.append('v1')
                    if len(depend_2.shape) == 1:
                        nontime_varying_depends.append('v2')
                elif depend_1 is not None:
                    tplot_data['v'] = depend_1
                    if len(depend_1.shape) == 1:
                        nontime_varying_depends.append('v')
                elif depend_2 is not None:
                    tplot_data['v2'] = depend_2
                    if len(depend_2.shape) == 1:
                        nontime_varying_depends.append('v')

                metadata[var_name] = {'display_type': var_atts.get("DISPLAY_TYPE", "time_series"),
                                      'scale_type': var_atts.get("SCALE_TYP"),
                                      'y_spec_scale_type': None,
                                      'var_attrs': var_atts,
                                      'labels': None,
                                      'file_name': filename,
                                      'global_attrs': gatt}

                labl_ptr = var_atts.get('LABL_PTR_1')
                if labl_ptr is not None:
                    try:
                        labl_ptr_arr = master_cdf_file.varget(labl_ptr)
                        if labl_ptr_arr is not None:
                            metadata[var_name]['labels'] = labl_ptr_arr.flatten().tolist()
                    except:
                        pass

                units = filter_greater_than(var_atts.get('UNITS'))
                if units is None:
                    unit_ptr = var_atts.get('UNIT_PTR')
                    if unit_ptr is not None:
                        try:
                            unit_ptr_array = master_cdf_file.varget(unit_ptr)
                            if unit_ptr_array is not None:
                                units = filter_greater_than(unit_ptr_array.flatten().tolist())
                        except:
                            pass
                metadata[var_name]['units'] = str(units)

                if metadata[var_name]['scale_type'] is None:
                    alt_scale_type = var_atts.get("SCALETYP", "linear")
                    if alt_scale_type is not None:
                        metadata[var_name]['scale_type'] = alt_scale_type

                # handle y-axis options for spectra
                if 'DEPEND_1' in var_atts:
                    if isinstance(var_atts['DEPEND_1'], str):
                        try:
                            depend_1_var_atts = master_cdf_file.varattsget(var_atts['DEPEND_1'])

                            scale_type = depend_1_var_atts.get('SCALETYP')
                            if scale_type is None:
                                scale_type = depend_1_var_atts.get('SCALE_TYP')

                            if scale_type is not None:
                                metadata[var_name]['y_spec_scale_type'] = scale_type

                            depend_1_units = depend_1_var_atts.get('UNITS')

                            if depend_1_units is not None:
                                metadata[var_name]['y_spec_units'] = depend_1_units
                                metadata[var_name]['DEPEND_1_UNITS'] = depend_1_units
                        except ValueError:
                            pass

                # options for multidimensional variables
                if 'DEPEND_2' in var_atts:
                    if isinstance(var_atts['DEPEND_2'], str):
                        try:
                            depend_2_var_atts = master_cdf_file.varattsget(var_atts['DEPEND_2'])
                            depend_2_units = depend_2_var_atts.get('UNITS')
                            if depend_2_units is not None:
                                metadata[var_name]['DEPEND_2_UNITS'] = depend_2_units
                        except ValueError:
                            # some variables aren't actually available
                            pass
                if 'DEPEND_3' in var_atts:
                    if isinstance(var_atts['DEPEND_3'], str):
                        try:
                            depend_3_var_atts = master_cdf_file.varattsget(var_atts['DEPEND_3'])
                            depend_3_units = depend_3_var_atts.get('UNITS')
                            if depend_3_units is not None:
                                metadata[var_name]['DEPEND_3_UNITS'] = depend_3_units
                        except ValueError:
                            # some variables aren't actually available
                            pass

                # Check if the variable already exists in the for loop output
                if var_name not in output_table:
                    output_table[var_name] = tplot_data
                else:
                    # If it does, loop though the existing variable's x,y,v,v2,v3,etc
                    var_data = output_table[var_name]
                    for output_var in var_data:
                        if output_var not in nontime_varying_depends:
                            if np.asarray(tplot_data[output_var]).ndim == 0 and np.equal(tplot_data[output_var], None):
                                # If there is nothing in the new variable, then pass
                                pass
                            elif np.asarray(var_data[output_var]).ndim == 0 and np.equal(var_data[output_var], None):
                                # If there is nothing in the old variable, then replace
                                var_data[output_var] = tplot_data[output_var]
                            else:  # If they both have something, then concatenate
                                var_data[output_var] = np.concatenate((var_data[output_var], tplot_data[output_var]))

    if notplot:
        return output_table

    for var_name in output_table.keys():
        to_merge = False
        if var_name in pyspedas.tplot_tools.data_quants.keys() and merge:
            prev_data_quant = pyspedas.tplot_tools.data_quants[var_name]
            to_merge = True

        try:
            attr_dict = {}
            if metadata.get(var_name) is not None:
                attr_dict["CDF"] = {}
                attr_dict["CDF"]["VATT"] = metadata[var_name]['var_attrs']
                attr_dict["CDF"]["GATT"] = metadata[var_name]['global_attrs']
                attr_dict["CDF"]["FILENAME"] = metadata[var_name]['file_name']
                attr_dict["CDF"]["LABELS"] = metadata[var_name]['labels']

                # populate data_att; used by PySPEDAS as a common interface to
                # data attributes such as units, coordinate system, etc
                attr_dict["data_att"] = {"coord_sys": "",
                                         #"units": metadata[var_name]['var_attrs'].get('UNITS'),
                                         "units": metadata[var_name]['units'],
                                         "depend_1_units": metadata[var_name].get('DEPEND_1_UNITS'),
                                         "depend_2_units": metadata[var_name].get('DEPEND_2_UNITS'),
                                         "depend_3_units": metadata[var_name].get('DEPEND_3_UNITS')}

                # populate depend_1_units in data_att, if it's not set
                if attr_dict['data_att']['depend_1_units'] is None and metadata[var_name]['var_attrs'].get('UNITS') is not None:
                    attr_dict['data_att']['depend_1_units'] = metadata[var_name]['var_attrs'].get('UNITS')

                # extract the coordinate system, if available
                vatt_keys = list(attr_dict["CDF"]["VATT"].keys())
                vatt_lower = [k.lower() for k in vatt_keys]
                if 'coordinate_system' in vatt_lower:
                    attr_dict['data_att']['coord_sys'] = filter_greater_than(
                        attr_dict["CDF"]["VATT"][vatt_keys[vatt_lower.index('coordinate_system')]])

                if 'labels' in vatt_lower:
                    if attr_dict["CDF"]["VATT"].get('labels') is not None:
                        if isinstance(attr_dict["CDF"]["VATT"]['labels'], str):
                            # check for line separators
                            # this fixes the legend for RBSP L3 EFW data
                            if '\\n' in attr_dict["CDF"]["VATT"]['labels']:
                                attr_dict["CDF"]["VATT"]['labels'] = attr_dict["CDF"]["VATT"]['labels'].split('\\n')
                            if '\\N' in attr_dict["CDF"]["VATT"]['labels']:
                                attr_dict["CDF"]["VATT"]['labels'] = attr_dict["CDF"]["VATT"]['labels'].split('\\N')
            store_data(var_name, data=output_table[var_name], attr_dict=attr_dict)
        except (TypeError, ValueError) as err:
            logging.warning("Exception of type %s raised during store_data call for variable %s", str(type(err)), var_name)
            logging.warning("Exception message: %s",str(err))
            continue

        if var_name not in stored_variables:
            stored_variables.append(var_name)

        if metadata.get(var_name) is not None:
            if metadata[var_name]['display_type'].lower() == "spectrogram":
                options(var_name, 'spec', 1)
            if metadata[var_name]['scale_type'] == 'log':
                if metadata[var_name]['display_type'].lower() == "spectrogram":
                    options(var_name, 'zlog', 1)
                else:
                    options(var_name, 'ylog', 1)
            if metadata[var_name].get('y_spec_scale_type') is not None:
                if metadata[var_name]['y_spec_scale_type'] == 'log':
                    options(var_name, 'ylog', 1)
            if metadata[var_name].get('y_spec_units') is not None:
                options(var_name, 'ysubtitle', '[' + metadata[var_name].get('y_spec_units') + ']')
            if metadata[var_name].get('var_attrs') is not None:
                if metadata[var_name]['var_attrs'].get('LABLAXIS') is not None:
                    options(var_name, 'ytitle', metadata[var_name]['var_attrs']['LABLAXIS'])
                if metadata[var_name]['var_attrs'].get('UNITS') is not None:
                    unitsstr = filter_greater_than(metadata[var_name]['var_attrs']['UNITS'])
                    if metadata[var_name]['display_type'].lower() == 'spectrogram':
                        options(var_name, 'ztitle', f'[{unitsstr}]')
                    else:
                        options(var_name, 'ysubtitle', f'[{unitsstr}]')

            # Gather up all options in the variable attribute section, toss them into options and see what sticks
            # JWL 2025-03-06: We want options() to warn about unrecognized option names.  If used here, it will
            # spam the logs with warnings.  There is also the possibility that someone could use an attribute name
            # for some other purpose, that collides with the PySPEDAS interpretation.
            #options(var_name, opt_dict=metadata[var_name]['var_attrs'])

        if to_merge is True:
            cur_data_quant = pyspedas.tplot_tools.data_quants[var_name]
            if isinstance(pyspedas.tplot_tools.data_quants[var_name], dict):  # non-record varying variable, shouldn't be merged
                continue
            plot_options = copy.deepcopy(pyspedas.tplot_tools.data_quants[var_name].attrs)
            pyspedas.tplot_tools.data_quants[var_name] = xr.concat([prev_data_quant, cur_data_quant], dim='time').sortby('time')
            pyspedas.tplot_tools.data_quants[var_name].attrs = plot_options

    if notplot:
        return output_table

    if plot:
        tplot(stored_variables)

    return stored_variables


def filter_greater_than_single(attr):
    """
    Returns any text to the left of > in a variable attribute
    (e.g., coordinate systems, units)
    Assumes input is a single value
    """
    if not isinstance(attr, str):
        return attr
    return attr.split('>')[0].rstrip()

def filter_greater_than(attr):
    """
    Strip CDF comments from attribute values (single value or array/list of values)
    Returns any text to the left of '>'
    Args:
        attr:

    Returns: str

    """
    if isinstance(attr,str):
        return filter_greater_than_single(attr)
    elif isinstance(attr,Iterable):
        return list(map(filter_greater_than_single,attr))
    else:
        return filter_greater_than_single(attr)


from pyspedas.utilities.datasets import find_datasets

# Refer to __init__.py for revision history before load routines split out into separate files

def datasets(instrument=None, label=True):
    return find_datasets(mission='CNOFS', instrument=instrument, label=label)


import os
import logging
import warnings
from shutil import copyfileobj, copy
from tempfile import NamedTemporaryFile
from pyspedas.tplot_tools import time_double, time_string
from pyspedas.projects.mms.mms_login_lasp import mms_login_lasp
from pyspedas.projects.mms.mms_config import CONFIG
from pyspedas.projects.mms.mec_ascii.mms_get_local_state_files import mms_get_local_state_files
from pyspedas.projects.mms.mec_ascii.mms_load_eph_tplot import mms_load_eph_tplot
from pyspedas.projects.mms.mec_ascii.mms_load_att_tplot import mms_load_att_tplot

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def mms_get_state_data(probe='1', trange=['2015-10-16', '2015-10-17'], 
    datatypes=['pos', 'vel'], level='def', no_download=False, pred_or_def=True, 
    suffix='', always_prompt=False):
    """
    Helper routine for loading state data (ASCII files from the SDC); not meant to be called directly; see pyspedas.projects.mms.state instead
    """

    if not isinstance(probe, list): probe = [probe]

    local_data_dir = CONFIG['local_data_dir']
    download_only = CONFIG['download_only']

    start_time = time_double(trange[0])-60*60*24.
    end_time = time_double(trange[1])

    # check if end date is anything other than 00:00:00, if so
    # add a day to the end time to ensure that all data is downloaded
    if type(trange[1]) == str:
        endtime_day = time_double(time_string(time_double(trange[1]), fmt='%Y-%m-%d'))
    else:
        endtime_day = time_double(time_string(trange[1], fmt='%Y-%m-%d'))

    if end_time > endtime_day:
        add_day = 60*60*24.
    else:
        add_day = 0

    start_time_str = time_string(start_time, fmt='%Y-%m-%d')
    end_time_str = time_string(end_time+add_day, fmt='%Y-%m-%d')

    filetypes = []

    if 'pos' in datatypes or 'vel' in datatypes:
        filetypes.append('eph')

    if 'spinras' in datatypes or 'spindec' in datatypes:
        filetypes.append('att')

    user = None
    if not no_download:
        sdc_session, user = mms_login_lasp(always_prompt=always_prompt)

    return_vars = []

    sep = "/" if is_fsspec_uri(local_data_dir) else os.path.sep

    for probe_id in probe:
        # probe will need to be a string from now on
        probe_id = str(probe_id)

        for filetype in filetypes:
            product = level + filetype

            files_in_interval = []
            out_files = []

            out_dir = sep.join([local_data_dir, 'ancillary', 'mms'+probe_id, level+filetype])

            if CONFIG['no_download'] != True and no_download != True:
                # predicted doesn't support start_date/end_date
                if level == 'def':
                    dates_for_query = '&start_date='+start_time_str+'&end_date='+end_time_str
                else:
                    dates_for_query = ''

                if user is None:
                    url = 'https://lasp.colorado.edu/mms/sdc/public/files/api/v1/file_info/ancillary?sc_id=mms'+probe_id+'&product='+product+dates_for_query
                else:
                    url = 'https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/file_info/ancillary?sc_id=mms'+probe_id+'&product='+product+dates_for_query

                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=ResourceWarning)
                    http_request = sdc_session.get(url, verify=True)
                    if http_request.status_code != 200:
                        logging.warning("Request to MMS SDC returned HTTP status code %d", http_request.status_code)
                        logging.warning("Text: %s", http_request.text)
                        logging.warning("URL: %s", url)
                        continue
                    else:
                        http_json = http_request.json()

                # since predicted doesn't support start_date/end_date, we'll need to parse the correct dates
                if level != 'def':
                    for file in http_json['files']:
                        # first, remove the dates that start after the end of the trange
                        if time_double(file['start_date']) > endtime_day:
                            continue
                        # now remove files that end before the start of the trange
                        if start_time > time_double(file['end_date']):
                            continue
                        files_in_interval.append(file)
                        break
                else:
                    files_in_interval = http_json['files']

                for file in files_in_interval:
                    out_file = sep.join([out_dir, file['file_name']])

                    if is_fsspec_uri(out_file):
                        protocol, path = out_file.split("://")
                        fs = fsspec.filesystem(protocol)

                        exists = fs.exists(out_file)
                    else:
                        exists = os.path.exists(out_file)

                    if exists:
                        if is_fsspec_uri(out_file):
                            f_size = fs.size(out_file)
                        else:
                            f_size = os.stat(out_file).st_size

                        if str(f_size) == str(file['file_size']):
                            out_files.append(out_file)
                            http_request.close()
                            continue

                    if user is None:
                        download_url = 'https://lasp.colorado.edu/mms/sdc/public/files/api/v1/download/ancillary?file=' + file['file_name']
                    else:
                        download_url = 'https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/download/ancillary?file=' + file['file_name']

                    logging.info('Downloading ' + file['file_name'] + ' to ' + out_dir)

                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore", category=ResourceWarning)
                        fsrc = sdc_session.get(download_url, stream=True, verify=True)

                    ftmp = NamedTemporaryFile(delete=False)

                    with open(ftmp.name, 'wb') as f:
                        copyfileobj(fsrc.raw, f)

                    if is_fsspec_uri(out_dir):
                        protocol, path = out_dir.split("://")
                        fs = fsspec.filesystem(protocol)

                        fs.makedirs(out_dir, exist_ok=True)

                        # if the download was successful, put at URI specified
                        fs.put(ftmp.name, out_file)
                    else:
                        if not os.path.exists(out_dir):
                            os.makedirs(out_dir)

                        # if the download was successful, copy to data directory
                        copy(ftmp.name, out_file)

                    out_files.append(out_file)
                    fsrc.close()
                    ftmp.close()

            if download_only:
                continue

            # if no files are found remotely, try locally
            if not out_files:
                out_files = mms_get_local_state_files(probe=probe_id, level=level, filetype=filetype, trange=[start_time_str, end_time_str])

            if filetype == 'eph':
                return_vars += mms_load_eph_tplot(sorted(out_files), level=level, probe=probe_id, datatypes=datatypes, suffix=suffix, trange=trange)
            elif filetype == 'att':
                return_vars += mms_load_att_tplot(sorted(out_files), level=level, probe=probe_id, datatypes=datatypes, suffix=suffix, trange=trange)

    if not no_download:
        sdc_session.close()

    return return_vars



import os
import numpy as np
from pyspedas.tplot_tools import time_double
import pandas as pd
import time
import logging
import pickle
import os


def read_data_files(out_files=None, dtype=None, out_type="np", save_pickle=False):
    """
    Read data on a daily basis with a 10-secs or other resolution.

    Parameters
    ----------
    out_files : list of str
        The list of the downloaded data file paths.
        This list can be obtained from the load() function, using downloadonly=True.
    dtype : str
        The data which will be read ('EICS' or 'SECS').
    out_type : str
        The return type: 'np' for numpy array, 'df' for pandas dataframe, 'dc' for dictionary.
        The default is 'np'.
    save_pickle : bool
        Whether to save the output as a pickle file.

    Returns
    -------
    Varies (np.ndarray, pd.DataFrame, or dict)

        The data read from the files in the specified format.
        If out_type is 'np', the data is a numpy array.
        If out_type is 'df', the data is a pandas dataframe.
        If out_type is 'dc', the data is a dictionary.
    """

    if out_files is None:
        logging.error("No data files provided.")
        return None

    dtype = dtype.upper()

    file_names_arr_Dir = out_files
    start_time = time.time()
    # Reading the data at each time stamp (per resolution secs) on one specific date.

    # input the data into one pd data frame. (four columns)
    if out_type == "df":
        if dtype == "EICS":
            colnames = ["latitude", "longitude", "Jx", "Jy"]
        if dtype == "SECS":
            colnames = ["latitude", "longitude", "J"]
        data_all = []
        for idx, file in enumerate(file_names_arr_Dir):
            df = pd.read_csv(file, header=None, sep=r"\s+", skiprows=0, names=colnames)
            df["datetime"] = file[-19:-4]
            data_all.append(df)
        output = pd.concat(data_all, axis=0, ignore_index=True)

    elif out_type == "np":
        latitude = []
        longitude = []
        date_time = []
        if dtype == "EICS":
            Jx = []
            Jy = []
            for file in file_names_arr_Dir:
                di = np.loadtxt(file)
                num_row = np.shape(di)[0]
                latitude.extend(di[:, 0])
                longitude.extend(di[:, 1])
                Jx.extend(di[:, 2])
                Jy.extend(di[:, 3])
                date_time.extend(np.full(num_row, file[-19:-4]))
            num_row2 = len(latitude)
            data_all = np.array([latitude, longitude, Jx, Jy, date_time])
            data_all = data_all.reshape([5, num_row2])
            data_all = np.transpose(data_all)

        if dtype == "SECS":
            J = []
            for file in file_names_arr_Dir:
                di = np.loadtxt(file)
                num_row = np.shape(di)[0]
                latitude.extend(di[:, 0])
                longitude.extend(di[:, 1])
                J.extend(di[:, 2])
                date_time.extend(np.full(num_row, file[-19:-4]))
            num_row2 = len(latitude)
            data_all = np.array([latitude, longitude, J, date_time])
            data_all = data_all.reshape([4, num_row2])
            data_all = np.transpose(data_all)

        output = data_all

    elif out_type == "dc":
        data_dict = {}
        Jx = []
        Jy = []
        J = []

        date_time = []
        flag = 0
        filename_day1 = file_names_arr_Dir[0]

        for idx, file in enumerate(
            file_names_arr_Dir
        ):  # per dat file with 1 min resolution.
            if not os.path.isfile(file):
                continue  # jump ouf of the current iteration, into the next iteration of the same loop.
            if os.stat(file).st_size == 0:  # check if the file is empty.
                continue

            di = np.loadtxt(file)
            if np.shape(di)[0] > 0 and flag == 0:
                num_row = np.shape(di)[0]  # np array
                latitude = di[:, 0]  # np array
                longitude = di[:, 1]  # np array
                flag = 1

            if dtype == "EICS":
                Jx.append(di[:, 2])  # list [np.arrays]
                Jy.append(di[:, 3])  # list [np.arrays]
            if dtype == "SECS":
                J.append(di[:, 2])  # list [np.arrays]
            date_time.append(file[-19:-4])  # list of str

        date_time = np.array(date_time)  # np array of str
        date_time = time_double(date_time)  # np array of float
        if dtype == "EICS":
            Jx = np.vstack(Jx)  # np array
            Jy = np.vstack(Jy)  # np array
            data_dict = {
                "time": date_time,
                "latitude": latitude,
                "longitude": longitude,
                "Jx": Jx,
                "Jy": Jy,
            }
        if dtype == "SECS":
            J = np.vstack(J)  # np array
            data_dict = {
                "time": date_time,
                "latitude": latitude,
                "longitude": longitude,
                "J": J,
            }
        output = data_dict

    else:
        logging.error("Invalid output type." + str(out_type))
        return None

    if save_pickle == True:
        if out_type == "dc":  # too large, not useful.
            with open("data_dc.pkl", "wb") as f:
                pickle.dump(output, f)

        # f.close()
    logging.info(
        "running time of output "
        + out_type
        + ": --- %s seconds ---" % (time.time() - start_time)
    )

    return output



import unittest
import logging
from pyspedas.projects.erg import mgf
from pyspedas.projects.erg.satellite.erg.common.cotrans.erg_cotrans import erg_cotrans
from pyspedas import tplot_names, get_data, data_exists, store_data
import numpy as np
from numpy.testing import assert_allclose



class MyTestCase(unittest.TestCase):

    # Note: the SGA and SGI axis directions in the attitude files are not necessarily reliable!
    # In this example, there are cases where SGA-X and SGI-Z are parallel or nearly so, when
    # they should be close to perpendicular.  This eventually leads to NaNs in the cotrans output.
    def test_cotrans(self):
        mgf_vars=mgf()
        # Clean NaNs from input data
        orig_time, orig_data = get_data('erg_mgf_l2_mag_8sec_dsi')
        orig_meta = get_data('erg_mgf_l2_mag_8sec_dsi',metadata=True)
        good_locations = np.where(np.isfinite(orig_data))
        good_times_idx = np.unique(good_locations[0])
        good_times = orig_time[good_times_idx]
        good_vals = orig_data[good_times_idx,:]
        store_data('erg_mgf_l2_mag_8sec_clean_dsi',{'x':good_times,'y':good_vals}, attr_dict=orig_meta)

        valid_coords=['dsi','sgi','sga','j2000']
        erg_cotrans('erg_mgf_l2_mag_8sec_clean_dsi', out_coord='sgi')
        dsi_sgi = get_data('erg_mgf_l2_mag_8sec_clean_sgi')
        sgi_nan = np.where(np.isnan(dsi_sgi[1]))
        self.assertTrue(len(sgi_nan[0]) == 0)
        erg_cotrans('erg_mgf_l2_mag_8sec_clean_dsi', out_coord='sga')
        dsi_sga = get_data('erg_mgf_l2_mag_8sec_clean_sga')
        sga_nan = np.where(np.isnan(dsi_sga[1]))
        # This turns out to have NaNs in it
        # self.assertTrue(len(sga_nan[0]) == 0)
        erg_cotrans('erg_mgf_l2_mag_8sec_clean_dsi', out_coord='j2000')
        self.assertTrue(data_exists('erg_mgf_l2_mag_8sec_clean_sgi'))
        self.assertTrue(data_exists('erg_mgf_l2_mag_8sec_clean_sga'))
        self.assertTrue(data_exists('erg_mgf_l2_mag_8sec_clean_j2000'))
        # Now test all combinations
        var_base = 'erg_mgf_l2_mag_8sec_clean'
        for in_coord in valid_coords:
            for out_coord in valid_coords:
                in_name = var_base + '_' + in_coord
                out_name = var_base + '_' + in_coord + '_' + out_coord
                erg_cotrans(in_name, out_name, in_coord=in_coord, out_coord=out_coord)
                self.assertTrue(data_exists(out_name))
        # Now test some round trips
        erg_cotrans('erg_mgf_l2_mag_8sec_clean_sgi', 'erg_mgf_l2_mag_8sec_clean_dsi_sgi_dsi', out_coord='dsi', )
        erg_cotrans('erg_mgf_l2_mag_8sec_clean_sga', 'erg_mgf_l2_mag_8sec_clean_dsi_sga_dsi', out_coord='dsi', )
        erg_cotrans('erg_mgf_l2_mag_8sec_clean_j2000', 'erg_mgf_l2_mag_8sec_clean_dsi_j2000_dsi', out_coord='dsi', )
        self.assertTrue(data_exists('erg_mgf_l2_mag_8sec_clean_dsi_sgi_dsi'))
        self.assertTrue(data_exists('erg_mgf_l2_mag_8sec_clean_dsi_sga_dsi'))
        self.assertTrue(data_exists('erg_mgf_l2_mag_8sec_clean_dsi_j2000_dsi'))
        orig_time, orig_data = get_data('erg_mgf_l2_mag_8sec_clean_dsi')
        sga_time, sga_data = get_data('erg_mgf_l2_mag_8sec_clean_dsi_sga_dsi')
        sgi_time, sgi_data = get_data('erg_mgf_l2_mag_8sec_clean_dsi_sgi_dsi')
        j2000_time, j2000_data = get_data('erg_mgf_l2_mag_8sec_clean_dsi_j2000_dsi')
        orig_nan = np.where(np.isnan(orig_data))
        sga_nan = np.where(np.isnan(sga_data))
        sgi_nan = np.where(np.isnan(sgi_data))
        j2000_nan = np.where(np.isnan(j2000_data))
        # This assertion fails due to presence of NaNs, and also some values look mismatched
        # assert_allclose(orig_data, sga_data, atol=0.1)
        assert_allclose(orig_data, sgi_data, atol=1e-06)
        assert_allclose(orig_data, j2000_data, atol=1e-06)


if __name__ == '__main__':
    unittest.main()



import re


def maven_kp_l2_regex():
    """
    Compile regular expressions for matching MAVEN KP and L2 file names.

    Parameters
    ----------

    Returns
    -------
    tuple 
        A tuple containing two compiled regular expressions:
            - kp_regex: Regular expression for matching Maven KP file names.
            - l2_regex: Regular expression for matching Maven L2 file names.
    """
    
    # kp pattern
    kp_pattern = (
        r"^mvn_(?P<{0}>kp)_"
        r"(?P<{1}>insitu|iuvs)"
        r"(?P<{2}>|_[a-zA-Z0-9\-]+)_"
        r"(?P<{3}>[0-9]{{4}})"
        r"(?P<{4}>[0-9]{{2}})"
        r"(?P<{5}>[0-9]{{2}})"
        r"(?P<{6}>|[t|T][0-9]{{6}})_"
        r"v(?P<{7}>[0-9]+)_r(?P<{8}>[0-9]+)\."
        r"(?P<{9}>tab)"
        r"(?P<{10}>\.gz)*"
    ).format(
        "instrument",
        "level",
        "description",
        "year",
        "month",
        "day",
        "time",
        "version",
        "revision",
        "extension",
        "gz",
    )

    kp_regex = re.compile(kp_pattern)

    l2_pattern = (
        r"^mvn_(?P<{0}>[a-zA-Z0-9]+)_"
        r"(?P<{1}>l[a-zA-Z0-9]+)"
        r"(?P<{2}>|_[a-zA-Z0-9\-]+)_"
        r"(?P<{3}>[0-9]{{4}})"
        r"(?P<{4}>[0-9]{{2}})"
        r"(?P<{5}>[0-9]{{2}})"
        r"(?P<{6}>|T[0-9]{{6}}|t[0-9]{{6}})_"
        r"v(?P<{7}>[0-9]+)_"
        r"r(?P<{8}>[0-9]+)\."
        r"(?P<{9}>cdf|xml|sts|md5)"
        r"(?P<{10}>\.gz)*"
    ).format(
        "instrument",
        "level",
        "description",
        "year",
        "month",
        "day",
        "time",
        "version",
        "revision",
        "extension",
        "gz",
    )
    l2_regex = re.compile(l2_pattern)
    return (kp_regex, l2_regex)



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot
import logging


from .config import CONFIG


def load(trange=['2018-11-5', '2018-11-6'],
         probe='a',
         instrument='emfisis',
         level='l3',
         datatype='magnetometer',
         prefix='',
         suffix='',
         force_download=False,
         cadence='4sec',  # for EMFISIS mag data
         coord='sm',  # for EMFISIS mag data
         wavetype='waveform',  # for EMFISIS waveform data
         rel='rel04',  # for ECT data
         get_support_data=False,
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    Load Van Allen Probes (RBSP) data for various instruments and data levels.

    It is not intended to be called directly. Instead, use the following wrappers for specific instruments:
        - pyspedas.projects.rbsp.emfisis
        - pyspedas.projects.rbsp.rbspice
        - pyspedas.projects.rbsp.efw
        - pyspedas.projects.rbsp.mageis
        - pyspedas.projects.rbsp.hope
        - pyspedas.projects.rbsp.rept
        - pyspedas.projects.rbsp.rps

    Parameters
    ----------
    trange : list of str
        Time range of interest in the format ['YYYY-MM-DD', 'YYYY-MM-DD'] or
        ['YYYY-MM-DD/hh:mm:ss', 'YYYY-MM-DD/hh:mm:ss']. Default is ['2018-11-5', '2018-11-6'].

    probe : str or list of str, default='a'
        Spacecraft probe name: 'a' or 'b'

    instrument : str, default='emfisis'
        Name of the instrument.

    level : str, default='l3'
        Data level

    datatype : str, default='magnetometer'
        Instrument-specific data type.

        Suffix for the tplot variable names.
    prefix : str, default=''
        Prefix for the tplot variable names.

    suffix : str, default=''
        Suffix for the tplot variable names.

    force_download : bool, default=False
        Download file even if local version is more recent than server version.

    cadence : str, default='4sec'
        Data cadence

    coord : str, default='sm'
        Data coordinate system.

    wavetype : str, default='waveform'
        Type of level 2 waveform data (applicable for WFR and HFR data).

    rel : str, default='rel04'
        Data release version (for ECT data).

    get_support_data : bool, default=False
        If True, loads data with attribute "VAR_TYPE" of "support_data".

    varformat : str, optional
        The file variable formats to load. Accepts wildcard character "*".

    varnames : list of str, optional
        List of variable names to load. If not specified, all variables are loaded.

    downloadonly : bool, default=False
        If True, downloads the CDF files but does not load them into tplot variables.

    notplot : bool, default=False
        If True, returns data in hash tables instead of creating tplot variables.

    no_update : bool, default=False
        If True, only load data from the local cache.

    time_clip : bool, default=False
        If True, clips the variables to the specified range in trange.

    Returns
    -------
    tvars : dict or list
        List of created tplot variables or dict of data tables if notplot is True.

    Examples
    --------
    This function is not intended to be called directly.
    """


    if not isinstance(probe, list):
        probe = [probe]

    datatype_in = datatype
    datatype = datatype.lower()
    out_files = []

    if notplot:
        tvars = {}
    else:
        tvars = []

    if prefix is None:
        prefix = ''

    if suffix is None:
        suffix = ''

    for prb in probe:
        if instrument == 'emfisis':
            if datatype in ['density', 'housekeeping', 'wna-survey']:
                pathformat = f'rbsp{prb}/{level}/{instrument}/{datatype}/%Y/rbsp-{prb}_{datatype}_{instrument}-{level}_%Y%m%d_v*.cdf'
            elif datatype in ['wfr', 'hfr']:
                pathformat = f'rbsp{prb}/{level}/{instrument}/{datatype}/{wavetype}/%Y/rbsp-{prb}_{datatype}-{wavetype}_{instrument}-{level}_%Y%m%d*_v*.cdf'
            else:
                if level == 'l2' and datatype == 'magnetometer':
                    pathformat = f'rbsp{prb}/{level}/{instrument}/{datatype}/uvw/%Y/rbsp-{prb}_{datatype}_uvw_{instrument}-{level}_%Y%m%d*_v*.cdf'
                else:
                    pathformat = f'rbsp{prb}/{level}/{instrument}/{datatype}/{cadence}/{coord}/%Y/rbsp-{prb}_{datatype}_{cadence}-{coord}_{instrument}-{level}_%Y%m%d_v*.cdf'
        elif instrument == 'rbspice':
            pathformat = f'rbsp{prb}/{level}/{instrument}/{datatype}/%Y/rbsp-{prb}-{instrument}_lev-{level[-1]}?{datatype}_%Y%m%d_v*.cdf'
            prefix = prefix + f'rbsp{prb}_rbspice_{level}_{datatype_in}_'
        elif instrument == 'efw':
                if level == 'l3':
                    pathformat = f'rbsp{prb}/{level}/{instrument}/%Y/rbsp{prb}_{instrument}-{level}_%Y%m%d_v??.cdf'
                else:
                    pathformat = f'rbsp{prb}/{level}/{instrument}/{datatype}/%Y/rbsp{prb}_{instrument}-{level}_{datatype}_%Y%m%d_v??.cdf'
        elif instrument == 'mageis':
                pathformat = f'rbsp{prb}/{level}/ect/{instrument}/sectors/{rel}/%Y/rbsp{prb}_{rel}_ect-mageis-{level}_%Y%m%d_v*.cdf'
        elif instrument == 'hope':
            if datatype == 'moments':
                pathformat = f'rbsp{prb}/{level}/ect/{instrument}/{datatype}/{rel}/%Y/rbsp{prb}_{rel}_ect-hope-mom-{level}_%Y%m%d_v*.cdf'
            elif datatype == 'pitchangle':
                pathformat = f'rbsp{prb}/{level}/ect/{instrument}/{datatype}/{rel}/%Y/rbsp{prb}_{rel}_ect-hope-pa-{level}_%Y%m%d_v*.cdf'
            elif datatype == 'spinaverage':
                pathformat = f'rbsp{prb}/{level}/ect/{instrument}/{datatype}/{rel}/%Y/rbsp{prb}_{rel}_ect-hope-sci-{level}sa_%Y%m%d_v*.cdf'
        elif instrument == 'rept':
            pathformat = f'rbsp{prb}/{level}/ect/{instrument}/sectors/{rel}/%Y/rbsp{prb}_{rel}_ect-rept-sci-{level}_%Y%m%d_v*.cdf'
        elif instrument == 'rps':
            if datatype == 'rps-1min':
                pathformat = f'rbsp{prb}/{level}/rps/psbr-rps-1min/%Y/rbsp{prb}_{level}-1min_psbr-rps_%Y%m%d_v*.cdf'
            else:
                pathformat = f'rbsp{prb}/{level}/rps/psbr-rps/%Y/rbsp{prb}_{level}_psbr-rps_%Y%m%d_v*.cdf'
        elif instrument == 'magephem':
            if cadence not in ['1min', '5min']:
                cadence = '1min'
                logging.info('Invalid cadence for magephem data. Defaulting to 1min.')
            if coord.lower() not in ['op77q', 't89d', 't89q', 'ts04d']:
                coord = 'op77q'
                logging.info('Invalid coordinate system for magephem data. Defaulting to op77q.')

            pathformat = f'rbsp{prb}/ephemeris/ect-mag-ephem/cdf/def-{cadence}-{coord}/%Y/rbsp-{prb}_mag-ephem_def-{cadence}-{coord}_%Y%m%d_v*.cdf'

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange)

        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'],
                         local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
        if files:
            out_files.extend(files)

        if not downloadonly:
            tvars_o = cdf_to_tplot(sorted(out_files), prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                                   varformat=varformat, varnames=varnames, notplot=notplot)

            if notplot:
                tvars.update(tvars_o)
            else:
                tvars.extend(tvars_o)

    if downloadonly:
        return sorted(out_files)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



from .load import load

# This routine was originally in polar/__init__.py.  If you need to see the history of this routine before
# it was moved to its own file, please check the history for __init__.py.

def cammice(trange=['2003-10-28', '2003-10-29'],
        datatype='k0',
        suffix='',  
        prefix='',
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        force_download=False):
    """
    This function loads data from the Charge and Mass Magnetospheric Ion Composition Experiment
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options:

        suffix: str
            The tplot variable names will be given this suffix. By default,
            no prefix is added.
            Default: ''
        
        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
        
        force_download: bool
            If True, downloads the file even if a newer version exists locally. 
            Default: False.

    Returns
    ----------
        List of tplot variables created.

    """
    return load(instrument='cammice', trange=trange, datatype=datatype, suffix=suffix, prefix=prefix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, force_download=force_download)




"""
This is a Python implementation of IDL's REBIN function.

https://www.nv5geospatialsoftware.com/docs/REBIN.html

It works for 1D, 2D, and 3D arrays.
"""

import numpy as np


def rebin1d(a, m, sample=False):
    """
    Resize an 1D array by interpolation or averaging.

    Parameters
    ----------
    array : array_like
        Input array to be resized.
    m : int
        New size for the output array.
    sample : bool, optional
        If True, use nearest-neighbor sampling (no interpolation/averaging).
        If False (default), use bilinear interpolation for expansion and averaging for compression.

    Returns
    -------
    ndarray
        Resized 1D array with the specified dimension.
        Type of array is the same as the input array.

    """
    a = np.asarray(a)
    if a.ndim != 1:
        raise ValueError("Input array must be 1D")

    if int(m) != m:
        raise ValueError("New size must be an integer")

    n = a.size  # Old size
    if n <= 0 or m <= 0:
        return np.array([])

    if m == 1:
        return np.array([np.mean(a)])

    if n == 1:
        return np.repeat(a, m)

    new_arr = np.zeros(m, dtype=a.dtype)  # keep the same dtype

    if m < n:
        # Downsample (Compression)
        if int(n / m) != n / m:
            raise ValueError("Compression to non-integer multiple is not supported")
        if sample:
            # Nearest neighbor sampling
            for i in range(m):
                p0 = int(i * n / m)
                if p0 >= n - 1:
                    # Repeat last value
                    new_arr[i] = a[n - 1]
                else:
                    new_arr[i] = a[p0]
        else:
            # Nearest neighbor averaging
            for i in range(m):
                new_arr[i] = np.mean(a[int(i * n / m) : int((i + 1) * n / m)])
    else:
        # Upsample (Expansion)
        if int(m / n) != m / n:
            raise ValueError("Expansion to non-integer multiple is not supported")
        if sample:
            # Nearest neighbor sampling
            for i in range(m):
                p0 = int(i * n / m)
                if p0 >= n - 1:
                    # Repeat last value
                    new_arr[i] = a[n - 1]
                else:
                    new_arr[i] = a[p0]
        else:
            # Bilinear interpolation
            for i in range(m):
                p = n * i / m
                if p >= n - 1:
                    # Repeat last value
                    new_arr[i] = a[n - 1]
                else:
                    p0 = int(p)
                    # To get the same results as in IDL when array a is integers:
                    # 1. Take the integer part only (not rounding), 3.667 -> 3
                    # 2. Perform the division last, so that we don't get 3.9997 which would become 3
                    new_arr[i] = a[p0] + (i * n - m * p0) * (a[p0 + 1] - a[p0]) / m

    return new_arr


def rebin(arr, new_dimensions, sample=False):
    """
    Resize an array by interpolation or averaging.

    Python implementation of IDL's REBIN function.

    Parameters
    ----------
    arr : array_like
        Input array to be resized.
    new_dimensions : int or tuple of ints
        New dimensions for the output array as a tuple of integers.
        For example, if new array is 2x3, pass (2, 3).
        New dimensions must be positive integers.
        Each dimension must be an integer multiple of the corresponding old dimension
            or the old dimension must be an integer multiple of the new dimension.
    sample : bool, optional
        If True, use nearest-neighbor sampling (no interpolation/averaging).
        If False (default), use bilinear interpolation for expansion and averaging for compression.

    Returns
    -------
    ndarray
        Resized array with the specified dimensions.
        If the input is array of int, the output will also be array of int.

    """
    arr = np.asarray(arr)
    old_shape = arr.shape
    if type(new_dimensions) is tuple:
        new_shape = new_dimensions
    else:
        new_shape = (new_dimensions,)

    # Validate dimensions
    if len(new_shape) != len(old_shape):
        raise ValueError(
            f"Number of dimensions must match: got {len(new_shape)}, expected {len(old_shape)}"
        )

    # Check if any dimension is zero or negative
    if any(d <= 0 for d in new_shape):
        raise ValueError("All dimensions must be positive")

    # Each new dimension has to be an integer multiple of the corresponding old dimension
    for old_dim, new_dim in zip(old_shape, new_shape):
        if old_dim < 1 or new_dim < 1 or (new_dim > old_dim and new_dim % old_dim != 0):
            raise ValueError(
                f"New dimension {new_dim} is not an integer multiple of old dimension {old_dim}"
            )
        elif new_dim < old_dim and old_dim % new_dim != 0:
            raise ValueError(
                f"Old dimension {old_dim} is not an integer multiple of new dimension {new_dim}"
            )

    # If shapes are the same, return a copy
    if new_shape == old_shape:
        return arr.copy()

    # Initialize new array, keep the same dtype
    new_arr = np.zeros(new_shape, dtype=arr.dtype)

    if len(new_shape) == 1:
        # One dimensional array
        m = new_shape[0]
        new_arr = rebin1d(arr, m, sample)

    elif len(new_shape) == 2:
        # Two dimensional array
        n1, n2 = old_shape
        m1, m2 = new_shape
        new_arr1 = np.zeros((m1, n2), dtype=arr.dtype)
        for i in range(n2):
            # Rebin each column
            new_arr1[:, i] = rebin1d(arr[:, i], m1, sample)

        new_arr = np.zeros((m1, m2), dtype=arr.dtype)
        for i in range(m1):
            # Rebin each row
            new_arr[i, :] = rebin1d(new_arr1[i, :], m2, sample)

    elif len(new_shape) == 3:
        # Three dimensional array
        n1, n2, n3 = old_shape
        m1, m2, m3 = new_shape

        # First rebin along dimension 0
        new_arr1 = np.zeros((m1, n2, n3), dtype=arr.dtype)
        for j in range(n2):
            for k in range(n3):
                new_arr1[:, j, k] = rebin1d(arr[:, j, k], m1, sample)

        # Then rebin along dimension 1
        new_arr2 = np.zeros((m1, m2, n3), dtype=arr.dtype)
        for i in range(m1):
            for k in range(n3):
                new_arr2[i, :, k] = rebin1d(new_arr1[i, :, k], m2, sample)

        # Finally rebin along dimension 2
        new_arr = np.zeros((m1, m2, m3), dtype=arr.dtype)
        for i in range(m1):
            for j in range(m2):
                new_arr[i, j, :] = rebin1d(new_arr2[i, j, :], m3, sample)

    else:
        raise ValueError("Only 1D, 2D, and 3D arrays are supported")

    return new_arr



from .load import load
from typing import List, Union, Optional

# This routine was moved out of __init__.py.  For preceding revision history, please see
# that file.


def staff(trange:List[str]=['2012-11-5', '2012-11-6'],
        probe:Union[str,List[str]]='1',
        datatype:str='pp',
        prefix:str='',
        suffix:str='',
        get_support_data:bool=False,
        varformat:str=None,
        varnames:List[str]=[],
        downloadonly:bool=False,
        notplot:bool=False,
        no_update:bool=False,
        time_clip:bool=False,
        force_download=False) -> List[str]:
    """
    Load data from the Cluster Spatio-Temporal Analysis of Field Fluctuation experiment
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2018-11-5', '2018-11-6']

        probe: list of str
            List of probes to load.  Valid options: '1','2','3','4'
            Default: '1'

        datatype: str
            Data type; Valid options:
            Default: 'pp'

        prefix: str
            The tplot variable names will be given this prefix.
            Default: ''

        suffix: str
            The tplot variable names will be given this suffix.
            Default: ''

        get_support_data: bool
            If True, Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted. If empty or None, all variables will be loaded.
            Default: None (all variables loaded)

        varnames: list of str
            List of CDF variable names to load (if empty or not specified,
            all data variables are loaded)
            Default: [] (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Returns
    -------
        list of str
            List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> staff_vars = pyspedas.projects.cluster.staff(trange=['2003-11-01','2003-11-02'],probe=['1','2'])
    >>> tplot(['B_par_f1__C1_PP_STA', 'B_perp_f1__C1_PP_STA', 'B_par_f1__C2_PP_STA', 'B_perp_f1__C2_PP_STA'])

    """
    return load(instrument='staff', trange=trange, probe=probe, datatype=datatype, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, no_update=no_update, time_clip=time_clip, force_download=force_download)





import logging
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(
    trange=["1996-12-01", "1996-12-02"],
    instrument="dcf",
    datatype="",
    level="l2",
    prefix="",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    Load FAST data into tplot variables.

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest [starttime, endtime] with the format
        ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
        ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Default is ["1996-12-01", "1996-12-02"].
    instrument : str or list of str, optional
        Type of instrument.
        Values can be: 'dcf', 'acf', 'esa', 'teams', 'all'.
        If 'all' is specified, all instruments will be loaded.
        Default is 'dcf'.
    datatype : str, optional
        Data type to load. Depends on the instrument.
        For 'esa' valid options are: 'eeb', 'ees', 'ieb', 'ies'.
        For all other insturments, this keyword is ignored.
        Default is ''.
    level : str, optional
        Data level to load. Depends on the instrument.
        For 'dcf' and 'teams' valid options are: 'l2', 'k0'.
        For all other instruments, this keyword is ignored.
        Default is 'l2'.
    prefix : str, optional
        The tplot variable names will be given this prefix.
        Default is ''.
        In all cases a suitable prefix will be given depending on the instrument.
    suffix : str, optional
        The tplot variable names will be given this suffix.
        Default is no suffix is added.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        Default is False; only loads in data with a "VAR_TYPE" attribute of "data".
    varformat : str, optional
        The file variable formats to load into tplot.
        Wildcard character "*" is accepted.
        Default is all variables are loaded in.
    varnames : list of str, optional
        List of variable names to load.
        Default is all data variables are loaded.
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables.
        Default is False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables.
        Default is False.
    no_update : bool, optional
        If set, only load data from your local cache.
        Default is False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword.
        Default is False.
    force_download : bool, optional
        Download file even if local version is more recent than server version.
        Default is False.

    Returns
    -------
    list of str/dictionary
        List of tplot variables created.
        If downloadonly is set to True, returns a list of the downloaded files.
        If notplot is set to True, returns a dictionary of the data loaded.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> dcf_vars = pyspedas.projects.fast.dcf(trange=["1996-12-01", "1996-12-02"])
    >>> tplot(['fast_dcf_DeltaB_GEI'])
    >>> acf_vars = pyspedas.projects.fast.acf(trange=["1996-12-01", "1996-12-02"])
    >>> tplot('fast_acf_HF_E_SPEC')
    >>> esa_vars = pyspedas.projects.fast.esa(trange=["1996-12-01", "1996-12-02"])
    >>> tplot('fast_esa_eflux')
    >>> teams_vars = pyspedas.projects.fast.teams(trange=["2005-08-01", "2005-08-02"])
    >>> tplot(['fast_teams_helium_omni_flux'])
    """

    out_files = []
    out_vars = []
    file_resolution = 24 * 3600.0

    if (
        trange is None
        or not isinstance(trange, list)
        or len(trange) != 2
        or trange[0] > trange[1]
    ):
        logging.error("Invalid trange specified.")
        return out_vars

    if not isinstance(instrument, list):
        instrument = [instrument]
    if "all" in instrument:
        instrument = ["dcf", "acf", "esa", "teams"]

    pathformat = ""
    for instr in instrument:
        if instr == "dcf":
            # levels are l2 (1996-1998) or k0 (1996-2002)
            if level == "k0":
                pathformat = "dcf/k0/%Y/fa_k0_dcf_%Y%m%d_v??.cdf"
            else:
                pathformat = "dcf/l2/dcb/%Y/%m/fast_hr_dcb_%Y%m%d%H????_?????_v??.cdf"
                file_resolution = 3600.0
        elif instr == "acf":
            # level k0 only (1996-2002)
            pathformat = "acf/k0/%Y/fa_k0_acf_%Y%m%d_v??.cdf"
        elif instr == "esa":
            # level l2 only
            # datatypes are eeb, ees, ieb, ies (1996-2009)
            if datatype not in ["eeb", "ees", "ieb", "ies"]:
                datatype = "eeb"  # default
            pathformat = (
                "esa/l2/"
                + datatype
                + "/%Y/%m/fa_esa_l2_"
                + datatype
                + "_%Y%m%d??????_*_v??.cdf"
            )
        elif instr == "teams":
            # levels are l2 (1996-2009) or k0 (1996-2009)
            if level == "k0":
                # no datatype for k0 data
                pathformat = "teams/k0/%Y/fa_k0_tms_%Y%m%d_v??.cdf"
            else:
                # for l2 data, the only available datatype is "pa"
                pathformat = "teams/l2/pa/%Y/%m/fast_teams_pa_l2_%Y%m%d_?????_v??.cdf"

        else:
            logging.error("Invalid instrument type: " + instr)
            continue

        # If prefix is not empty, add it to the pre variable
        pre = "fast_" + instr + "_"
        if prefix != "":
            pre = prefix + pre

        # find the full remote path names using the trange
        remote_names = dailynames(
            file_format=pathformat, trange=trange, res=file_resolution
        )

        # download the files
        files = download(
            remote_file=remote_names,
            remote_path=CONFIG["remote_data_dir"],
            local_path=CONFIG["local_data_dir"],
            no_download=no_update,
            force_download=force_download,
        )
        if files is not None:
            if not isinstance(files, list):
                files = [files]
            out_files.extend(files)

        if not downloadonly and len(files) > 0:
            # Read the files into tplot variables
            vars = cdf_to_tplot(
                files,
                prefix=pre,
                suffix=suffix,
                get_support_data=get_support_data,
                varformat=varformat,
                varnames=varnames,
                notplot=notplot,
            )
            if not isinstance(vars, list):
                vars = [vars]
            out_vars.extend(vars)

    out_files = list(set(out_files))
    out_files = sorted(out_files)

    if not downloadonly and len(out_files) < 1:
        logging.info("No files were downloaded.")
        return out_files  # return an empty list

    if downloadonly:
        return out_files

    if notplot:
        if len(out_vars) < 1:
            logging.info("No variables were loaded.")
            return {}
        else:
            return out_vars[0]  # return data in hash tables

    if time_clip:
        tclip(out_vars, trange[0], trange[1], suffix="", overwrite=True)

    return out_vars




from .load import load
from pyspedas.tplot_tools import options, data_quants

# This routine was originally in psp/__init__.py.  If you need to see the history of this routine before
# it was moved to its own file, please check the history for __init__.py.

# Loading
def fields(trange=['2018-11-5', '2018-11-6'], 
        datatype='mag_rtn', 
        level='l2',
        suffix='',  
        prefix='',
        get_support_data=False, 
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=False,
        username=None,
        password=None,
        last_version=False,
        force_download=False,
        version=None,
        ):
    """
    This function loads Parker Solar Probe FIELDS data
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        datatype: str
            Data type; Valid options include:
                'mag_RTN'
                'mag_RTN_1min'
                'mag_rtn_4_per_cycle' (SPDF only)
                'mag_RTN_4_Sa_per_Cyc' 
                'mag_SC'
                'mag_SC_1min'
                'mag_sc_4_per_cycle' (SPDF only)
                'mag_SC_4_Sa_per_Cyc' 
                'mag_VSO' (limited dates)
                'rfs_burst' (limited dates)
                'rfs_hfr', 
                'rfs_lfr'
                'f2_100bps'
                'dfb_dc_spec'
                'dfb_ac_spec'
                'dfb_dc_xspec'
                'dfb_ac_xspec'
                'merged_scam_wf'
                'sqtn_rfs_V1V2'

        suffix: str
            The tplot variable names will be given this suffix. By default,
            no prefix is added.
            Default: ''

        prefix: str
            The tplot variable names will be given this prefix.  By default,
            no prefix is added.
            Default: ''

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, this flag is False but 
            FIELDS support data is always loaded for datatypes where filtering
            on quality flags is supported.

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into 
            tplot variables

        notplot: bool
            Return the data in hash tables instead of creating tplot variables

        no_update: bool
            If set, only load data from your local cache

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword

        username: str
            Username to use for authentication.

            If passed, attempt to download data from the FIELDS Instrument Team server
            instead of the fully public server at SPDF.
            Provides access to unpublished, V02 files.

            Implemented for dataypes:
                'mag_RTN_1min'
                'mag_RTN_4_Sa_per_Cyc'
                'mag_SC'
                'mag_SC_1min'
                'mag_SC_4_Sa_per_Cyc'
                'sqtn_rfs_V1V2'

        password: str
            Password to use for authentication

        last_version: bool
            If True, only download the highest-numbered file version

        force_download: bool
            If True, downloads the file even if a newer version exists locally. 
            Default: False.

        version: str
            If supplied, the load routine will look for this specific CDF version, rather than a wildcard pattern
            matching any version.  Valid version strings look like "vN.M" for FIELDS sqtn_rfs_v1v1 data, or
            "vNN" for all other data types. Default: None

    Returns
    ----------
        List of tplot variables created.

    """
    if suffix is None:
        suffix = ''

    if prefix is None:
        prefix = ''    

    # SCaM and QTN data are Level 3
    if datatype.lower() in ['merged_scam_wf', 'sqtn_rfs_v1v2']:
        level = 'l3'
        print("Using LEVEL=L3")

    spec_types = None
    if datatype in ['dfb_dc_spec', 'dfb_ac_spec', 'dfb_dc_xspec', 'dfb_ac_xspec']:
        if level == 'l1':
            spec_types = ['1', '2', '3', '4']
        else:
            if datatype == 'dfb_dc_spec' or datatype == 'dfb_ac_spec':
                spec_types = ['dV12hg','dV34hg','dV12lg','dV34lg',
                    'SCMulfhg','SCMvlfhg','SCMwlfhg',
                    'SCMulflg','SCMvlflg','SCMwlflg',
                    'SCMdlfhg','SCMelfhg','SCMflfhg',
                    'SCMdlflg','SCMelflg','SCMflflg',
                    'SCMmf', 'V5hg']
            else:
                spec_types = ['SCMdlfhg_SCMelfhg','SCMdlfhg_SCMflfhg','SCMelfhg_SCMflfhg',
                    'SCMulfhg_SCMvlfhg','SCMulfhg_SCMwlfhg','SCMvlfhg_SCMwlfhg',
                    'dV12hg_dV34hg']

    loaded_vars = load(
        instrument='fields', trange=trange, datatype=datatype, spec_types=spec_types, level=level, 
        suffix=suffix, prefix=prefix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, 
        downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update,
        username=username, password=password,last_version=last_version, force_download=force_download, version=version,
    )
    
    if loaded_vars is None or notplot or downloadonly:
        return loaded_vars

    qf_root = prefix + 'psp_fld_' + level + '_quality_flags'+suffix

    # If variables are loaded that quality flag filtering supports --
    # Make sure the quality flag variable is also loaded and linked. 
    mag_rtnvars = [x for x in loaded_vars if 'fld_l2_mag_RTN' in x ]
    mag_scvars = [x for x in loaded_vars if 'fld_l2_mag_SC' in x ]
    rfs_vars = [x for x in loaded_vars if 'rfs_lfr' in x or 'rfs_hfr' in x]
    if (len(mag_rtnvars + mag_scvars + rfs_vars) > 0) \
        & ('psp_fld_' + level + '_quality_flags'+suffix not in loaded_vars):
        loaded_extra = load(
            instrument='fields', trange=trange, datatype=datatype, spec_types=spec_types, level=level, 
            suffix=suffix, prefix=prefix, get_support_data=True, varformat=varformat, varnames=['psp_fld_' + level + '_quality_flags'],
            downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update,
            username=username, password=password,last_version=last_version, force_download=force_download, version=version,
        )
        qf_root = prefix+'psp_fld_' + level + '_quality_flags'+suffix if prefix+'psp_fld_' + level + '_quality_flags'+suffix in loaded_extra else None
        loaded_vars += loaded_extra

    for var in mag_rtnvars:
        options(var, 'legend_names', ['Br (RTN)', 'Bt (RTN)', 'Bn (RTN)'])
        data_quants[var] = data_quants[var].assign_attrs({'qf_root':qf_root})

    for var in mag_scvars:
        options(var, 'legend_names', ['Bx', 'By', 'Bz'])
        data_quants[var] = data_quants[var].assign_attrs({'qf_root':qf_root})

    for var in rfs_vars:
        data_quants[var] = data_quants[var].assign_attrs({'qf_root':qf_root})

    return loaded_vars





import os
import unittest
import pyspedas
from pyspedas.tplot_tools import tplot, get_data
import numpy as np
from pyspedas.utilities.config_testing import TESTING_CONFIG

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
analysis_dir = "analysis_tools"
save_dir = os.path.join(output_dir, analysis_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


class MagNullTestCases(unittest.TestCase):

    def test_null_classification_x(self):
        # Test direct detection of 2-D type X nulls. (This will seldom or never occur in practice,
        # but see below for some degenerate cases that do appear in real data.
        l1 = complex(0.0, 0.0)
        l2 = complex(-0.5, 0.0)
        l3 = complex(0.5, 0.0)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 1)

    def test_null_classification_o(self):
        # Test direct detection of 2-D type O nulls.  Again, this is not expected in practice, except
        # for degenerate cases checked below.
        l1 = complex(0.0, 0.0)
        l2 = complex(0.0, 1.0)
        l3 = complex(0.0, -1.0)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 2)

    def test_null_classification_a(self):
        # Test for 3-D type A (radial) null
        l1 = complex(-0.25, 0.0)
        l2 = complex(-0.25, 0.0)
        l3 = complex(0.5, 0.0)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 3)

    def test_null_classification_b(self):
        # Test for 3-D type B (radial) null
        l1 = complex(0.25, 0.0)
        l2 = complex(0.25, 0.0)
        l3 = complex(-0.5, 0.0)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 4)

    def test_null_classification_as(self):
        # Test for 3-D type A_s (spiral) null
        l1 = complex(0.5, 0.0)
        l2 = complex(-0.25, 0.25)
        l3 = complex(-0.25, -0.25)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 5)

    def test_null_classification_bs(self):
        # Test for 3-D type B_s (spiral) null
        l1 = complex(-0.5, 0.0)
        l2 = complex(0.25, 0.25)
        l3 = complex(0.25, -0.25)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 6)

    def test_null_classification_a_degen(self):
        # Test for type A null, degenerating to type X if the ratio of mim to max norm is < 0.25
        l1 = complex(-0.01, 0.0)
        l2 = complex(-0.1, 0.0)
        l3 = complex(0.89, 0.0)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 7)

    def test_null_classification_b_degen(self):
        # Test for type B null, degenerating to type X if the ratio of min to max norm is < 0.25
        l1 = complex(0.01, 0.0)
        l2 = complex(0.1, 0.0)
        l3 = complex(-0.89, 0.0)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 8)

    def test_null_classification_as_degen(self):
        # Test for type A_s null, degenerating to type O if the ratio of max(abs(real)) to min(abs(imag)) is < 0.25
        l1 = complex(0.1, 0.0)
        l2 = complex(-0.1, 0.5)
        l3 = complex(-0.1, -0.5)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 9)

    def test_null_classification_bs_degen(self):
        # Test for type B_s null, degenerating to type O if the ratio of max(abs(real)) to min(abs(imag)) is < 0.25
        l1 = complex(-0.1, 0.0)
        l2 = complex(-0.1, 0.5)
        l3 = complex(0.1, -0.5)
        lambdas = [l1, l2, l3]
        tc = pyspedas.classify_null_type(lambdas)
        self.assertEqual(tc, 10)

    def test_find_magnetic_nulls_fote_mms(self):
        data = pyspedas.projects.mms.fgm(
            probe=[1, 2, 3, 4],
            trange=["2016-09-19/07:40", "2016-09-19/07:45"],
            data_rate="srvy",
            time_clip=True,
            varformat="*_gse_*",
            get_fgm_ephemeris=True,
        )
        self.assertEqual(len(data), 16)
        fields = ["mms" + prb + "_fgm_b_gse_srvy_l2" for prb in ["1", "2", "3", "4"]]
        positions = ["mms" + prb + "_fgm_r_gse_srvy_l2" for prb in ["1", "2", "3", "4"]]
        null_vars = pyspedas.find_magnetic_nulls_fote(
            fields=fields,
            positions=positions,
            smooth_fields=True,
            smooth_npts=10,
            smooth_median=True,
        )
        local_png = os.path.join(save_dir, "mms_null_vars.png")
        tplot(null_vars, save_png=local_png, display=global_display)
        # We reconstruct the field vectors at the s/c positions using the Jacobian and the field value at the tetrahedron
        # barycenter.  The max difference between the observations and the reconstructed values for each time step
        # should be extremely close to zero, so that's what we'll use for our test.
        d = get_data("max_reconstruction_error")
        np.testing.assert_allclose(d.y, 0.0, atol=1.0e-10)

    def test_find_magnetic_nulls_fote_cluster(self):
        data = pyspedas.projects.cluster.load_csa(
            probes=["C1", "C2", "C3", "C4"],
            trange=["2003-08-17/16:40", "2003-08-17/16:45"],
            datatypes="CP_FGM_FULL",
            time_clip=True,
        )
        self.assertEqual(len(data), 12)
        # tplot_names()
        fields = [
            "B_vec_xyz_gse__C" + prb + "_CP_FGM_FULL" for prb in ["1", "2", "3", "4"]
        ]
        positions = [
            "sc_pos_xyz_gse__C" + prb + "_CP_FGM_FULL" for prb in ["1", "2", "3", "4"]
        ]
        null_vars = pyspedas.find_magnetic_nulls_fote(
            fields=fields, positions=positions, smooth_fields=True
        )
        local_png = os.path.join(save_dir, "cluster_null_vars.png")
        tplot(null_vars, display=global_display, save_png=local_png)
        d = get_data("max_reconstruction_error")
        np.testing.assert_allclose(d.y, 0.0, atol=1.0e-10)


if __name__ == "__main__":
    unittest.main()



from pyspedas.tplot_tools import tplot
import pyspedas
from .load import load

# This routine was originally in swarm/__init__.py.
def mag(trange=['2017-03-27/06:00', '2017-03-27/08:00'],
        probe='a',
        datatype='hr',
        level='l1b',
        prefix='',
        suffix='',
        varnames=[],
        time_clip=False,
        force_download=False):
    """
    Loads data from the Vector Field Magnetometer (VFM).

    Parameters
    ----------
    trange : list of str, default=['2017-03-27/06:00', '2017-03-27/08:00']
        Time range of interest [starttime, endtime] with the format
        'YYYY-MM-DD' or 'YYYY-MM-DD/hh:mm:ss'.
    probe : str or list of str, default='a'
        Swarm spacecraft ID ('a', 'b', and/or 'c').
    datatype : str, default='hr'
        Data type; valid options: 'hr' (high resolution), 'lr' (low resolution).
    level : str, default='l1b'
        Data level; options: 'l1b'.
    prefix : str, optional
        The tplot variable names will be given this prefix. By default, no prefix is added.
    suffix : str, optional
        The tplot variable names will be given this suffix. By default, no suffix is added.
    varnames : list of str, optional
        List of variable names to load. If not specified, all data variables are loaded.
    time_clip : bool, default=False
        Time clip the variables to exactly the range specified in the trange keyword.
    force_download : bool, default=False
        Swarm data is requested via HAPI. This parameter always ignored and reserved for compatibility.

    Returns
    -------
    out_vars : list of str
        List of tplot variables created.

    Examples
    --------
    To load and plot Magnetometer (MAG) data from the Swarm mission for probe 'c' over a specific time range, you can use the following commands:

    >>> import pyspedas
    >>> from pyspedas import tplot

    # Load MAG data for probe 'c'
    >>> mag_vars = pyspedas.projects.swarm.mag(probe='c', trange=['2017-03-27/06:00', '2017-03-27/08:00'], datatype='hr')

    # Plot the loaded MAG data
    >>> tplot('swarmc_B_VFM')
    """

    return load(instrument='mag', trange=trange, probe=probe, level=level, datatype=datatype, prefix=prefix, suffix=suffix, varnames=varnames, time_clip=time_clip, force_download=force_download)
